/*!   
    \file WSControl.js
    \brief WebSocket Control Class
    \author zafaco GmbH <info@zafaco.de>
    \date Last update: 2017-11-24

    Copyright (C) 2016 - 2017 zafaco GmbH

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3 
    as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

var jsTool = new JSTool();




/*-------------------------class WSControl------------------------*/

/**
 * @class WSControl
 * @description WebSocket Control Class
 */
function WSControl()
{
    this.wsMeasurement  = '';
    
    var wsTestCase;
    var wsData;
    var wsDataTotal;
    var wsFrames;
    var wsFramesTotal;
    var wsSpeedAvgBitS;
    var wsSpeedAvgMBitS;
    var wsSpeedMinBitS;
    var wsSpeedMaxBitS;
    var wsOverhead;
    var wsOverheadTotal;
    var wsStartTime;
    var wsEndTime;
    var wsMeasurementTime;
    var wsMeasurementTimeTotal;
    var wsStartupStartTime;
    var wsOverheadPerFrame;
    var wsWorkers;
    var wsWorkersStatus;
    var wsWorkerTime;
    var wsInterval;
    var wsMeasurementRunningTime;
    var wsCompleted;
    var wsReportInterval;
    var wsTimeoutTimer;
    var wsTimeout;
    var wsStartupTimeout;
    var wsProtocol;
    var wsStreamsStart;
    var wsStreamsEnd;
    var wsFrameSize;
    var wsMeasurementError;
    
    
    var wsStateConnecting           = 0;
    var wsStateOpen                 = 1;
    var wsStateClosing              = 2;
    var wsStateClosed               = 3;
    
    
    var rttReportInterval           = 501;
    var rttProtocol                 = 'wsRoundTripTimeProtocol';
    
   
    var dlReportInterval            = 501;
    var dlWsOverheadPerFrame        = 4;
    var dlData                      = 0;
    var dlFrames                    = 0;
    var dlStartupData               = 0;
    var dlStartupFrames             = 0;
    var dlProtocol                  = 'wsDownloadProtocol';
    
    
    var ulReportInterval            = 501;
    var ulWsOverheadPerFrame        = 8;
    var ulReportDict                = {};
    var ulStartupData               = 0;
    var ulStartupFrames             = 0;
    var ulProtocol                  = 'wsUploadProtocol';


    var wsTarget;
    var wsRttTarget;
    var wsTargetPort;
    
    //default measurement parameters
    var rttRequests                 = 10;
    var rttRequestTimeout           = 1000;
    var rttRequestWait              = 500;
    var rttTimeout                  = (rttRequests * (rttRequestTimeout + rttRequestWait)) * 1.1;
    var rttPayloadSize              = 64;
    
    var dlStartupTime               = 3000;
    var dlMeasurementRunningTime    = 10000;
    var dlParallelStreams           = 4;
    var dlTimeout                   = 15000;
    var dlFrameSize                 = 32768;
    
    var ulStartupTime               = 3000;
    var ulMeasurementRunningTime    = 10000;
    var ulParallelStreams           = 4;
    var ulTimeout                   = 15000;
    var ulFrameSize                 = 65500;
    
    var wsParallelStreams;
    var wsStartupTime;
    
    var singleThread                = false;
    
    var wsWorkerPath                = 'WSWorker.js';




    /*-------------------------KPIs------------------------*/
    
    //Availability KPIs
    var wsErrorCode;
    var wsErrorDescription;
    
    
    //Round Trip Time KPIs
    var wsRttAvg;
    var wsRttMed;
    var wsRttMin;
    var wsRttMax;
    var wsRttRequests;
    var wsRttReplies;
    var wsRttErrors;
    var wsRttMissing;
    var wsRttPacketsize;


    //Download KPIs
    var wsDownloadRateAvg;
    var wsDownloadRateMin;
    var wsDownloadRateMax;
    var wsDownloadData;
    var wsDownloadDataTotal;
    var wsDownloadDuration;
    var wsDownloadDurationTotal;
    var wsDownloadStreamsStart;
    var wsDownloadStreamsEnd;
    var wsDownloadFrameSize;
    var wsDownloadFrames;
    var wsDownloadFramesTotal;
    var wsDownloadOverheadPerFrame;
    var wsDownloadOverhead;
    var wsDownloadOverheadTotal;
    var wsDownloadRateAvgMBitS;
    

    //Upload KPIs
    var wsUploadRateAvg;
    var wsUploadRateMin;
    var wsUploadRateMax;
    var wsUploadData;
    var wsUploadDataTotal;
    var wsUploadDuration;
    var wsUploadDurationTotal;
    var wsUploadStreamsStart;
    var wsUploadStreamsEnd;
    var wsUploadFrameSize;
    var wsUploadFrames;
    var wsUploadFramesTotal;
    var wsUploadOverheadPerFrame;
    var wsUploadOverhead;
    var wsUploadOverheadTotal;
    var wsUploadRateAvgMBitS;
    
    
    
    
    /*-------------------------public functions------------------------*/
    
    /**
     * @function measurementSetup
     * @description Function to Start test cases
     * @public
     * @param {string} measurementParameters JSON coded measurement Parameters
     */
    this.measurementSetup = function(measurementParameters)
    {
        resetValues();
        
        wsMeasurementError      = false;
        measurementParameters   = JSON.parse(measurementParameters);
        wsTestCase              = measurementParameters.testCase;

        switch (wsTestCase)
        {
            case 'rtt':
            {
                if (typeof measurementParameters.wsRttRequests !== 'undefined')        rttRequests         = Number(measurementParameters.wsRttRequests);
                if (typeof measurementParameters.wsRttRequestTimeout !== 'undefined')  rttRequestTimeout   = Number(measurementParameters.wsRttRequestTimeout);
                if (typeof measurementParameters.wsRttRequestWait !== 'undefined')     rttRequestWait      = Number(measurementParameters.wsRttRequestWait);
                if (typeof measurementParameters.wsRttTimeout !== 'undefined')         rttTimeout          = Number(measurementParameters.wsRttTimeout); 
                if (typeof measurementParameters.wsRttPayloadSize !== 'undefined')     rttPayloadSize      = Number(measurementParameters.wsRttPayloadSize);
                wsParallelStreams                                                   = 1;
                wsMeasurementRunningTime                                            = rttTimeout;
                wsReportInterval                                                    = rttReportInterval;
                wsProtocol                                                          = rttProtocol;
                wsTimeout                                                           = dlTimeout;
                break;
            }
            case 'download':
            {
                wsOverheadPerFrame          = dlWsOverheadPerFrame;
                wsParallelStreams           = dlParallelStreams;
                wsStartupTime               = dlStartupTime;
                wsMeasurementRunningTime    = dlMeasurementRunningTime;
                wsReportInterval            = dlReportInterval;
                wsTimeout                   = dlTimeout;
                wsProtocol                  = dlProtocol;
                wsFrameSize                 = dlFrameSize;
                break;
            }
            case 'upload':
            {
                wsFrameSize                 = ulFrameSize;
                
                if (ulFrameSize >= 65536)
                {
                    ulWsOverheadPerFrame    = 12;
                }
                else if (ulFrameSize < 126)
                {
                    ulWsOverheadPerFrame    = 6;
                }
                
                wsOverheadPerFrame          = ulWsOverheadPerFrame;
                wsUploadOverheadPerFrame    = ulWsOverheadPerFrame;
                wsParallelStreams           = ulParallelStreams;
                wsStartupTime               = ulStartupTime;
                wsMeasurementRunningTime    = ulMeasurementRunningTime;
                wsReportInterval            = ulReportInterval;
                wsTimeout                   = ulTimeout;
                wsProtocol                  = ulProtocol;
                break;
            }
        }
        
        if (typeof measurementParameters.wsTarget !== 'undefined')             wsTarget                    = String(measurementParameters.wsTarget);
        if (typeof measurementParameters.wsRttTarget !== 'undefined')          wsRttTarget                 = String(measurementParameters.wsRttTarget);        
        if (typeof measurementParameters.wsTargetPort !== 'undefined')         wsTargetPort                = String(measurementParameters.wsTargetPort);
        if (typeof measurementParameters.wsStartupTime !== 'undefined')        wsStartupTime               = Number(measurementParameters.wsStartupTime);
        if (typeof measurementParameters.wsTimeout !== 'undefined')            wsTimeout                   = Number(measurementParameters.wsTimeout);
        if (typeof measurementParameters.wsParallelStreams  !== 'undefined' && (wsTestCase === 'download' || wsTestCase === 'upload')) wsParallelStreams           = Number(measurementParameters.wsParallelStreams);
        if (typeof measurementParameters.wsMeasureTime      !== 'undefined' && (wsTestCase === 'download' || wsTestCase === 'upload')) wsMeasurementRunningTime    = Number(measurementParameters.wsMeasureTime);
        
        reportToMeasurement('info', 'starting measurement');
        
        console.log(wsTestCase + ': starting measurement using parameters:');
        var wsWssString                 = 'ws://';
        
        if (wsTestCase === 'rtt')
        {
            console.log('target:            ' + wsWssString + wsRttTarget + ':' + wsTargetPort);
            console.log('protocol:          ' + rttProtocol);
            console.log('requests:          ' + rttRequests);
            console.log('request timeout:   ' + rttRequestTimeout);
            console.log('request wait:      ' + rttRequestWait);
            console.log('timeout:           ' + rttTimeout);
        }
        
        if (wsTestCase === 'download' || wsTestCase === 'upload')
        {
            console.log('target:            ' + wsWssString + wsTarget + ':' + wsTargetPort);
            console.log('protocol:          ' + wsProtocol);
            console.log('startup time:      ' + wsStartupTime);
            console.log('measurement time:  ' + wsMeasurementRunningTime);
            console.log('parallel streams:  ' + wsParallelStreams);
            console.log('timeout:           ' + wsTimeout);        
        }
        
        wsWorkers       = new Array(wsParallelStreams);
        wsWorkersStatus = new Array(wsParallelStreams);

        for (var wsID = 0; wsID < wsWorkers.length; wsID++)
        {
            if (wsTestCase === 'upload')
            {
                ulReportDict[wsID]  = {};
            }
            
            var workerData = prepareWorkerData('connect', wsID);
            
            if (typeof measurementParameters.singleThread !== 'undefined')
            {
                singleThread                    = true;
                delete(wsWorkers[wsID]);
                wsWorkers[wsID]                 = new WSWorker();
                wsWorkers[wsID].wsControl       = this;
                setTimeout(wsWorkers[wsID].onmessage, 100,  workerData);
            }
            else
            {
                wsWorkersStatus[wsID]   = wsStateClosed;
                wsWorkers[wsID]         = new Worker(wsWorkerPath);
                wsWorkers[wsID].addEventListener('message', function (event)
                {
                    workerCallback(JSON.parse(event.data)); 
                }, false);
                
                wsWorkers[wsID].postMessage(workerData);
            }
        }
        
        wsTimeoutTimer = setTimeout(measurementTimeout, wsTimeout);  
    };

    /**
     * @function workerCallback
     * @description Function to receive callbacks from the WSWorkers
     * @public
     * @param {string} data JSON coded measurement Results
     */
    this.workerCallback = function(data)
    {
        workerCallback(JSON.parse(data));
    };
    
    
    
    
    /*-------------------------private functions------------------------*/
    
    /**
     * @function workerCallback
     * @description Function to receive callbacks from the WSWorkers
     * @private
     * @param {string} data measurement Results
     */
    function workerCallback(data)
    {
        switch (data.cmd)
        {
            case 'info':
            {
                wsWorkersStatus[data.wsID] = data.wsState;
                console.log('wsWorker ' + data.wsID + ' command: \'' + data.cmd + '\' message: \'' + data.msg);
                
                break;
            }

            case 'open':
            {
                wsWorkersStatus[data.wsID] = data.wsState;             
                var allOpen = true;

                for (var wsWorkersStatusID = 0; wsWorkersStatusID < wsWorkersStatus.length; wsWorkersStatusID++)
                {
                    if (wsWorkersStatus[wsWorkersStatusID] !== wsStateOpen)
                    {
                        allOpen = false;
                    }
                }

                console.log('wsWorker ' + data.wsID + ' websocket open');

                if (allOpen)
                {
                    console.log('all websockets open');

                    clearTimeout(wsTimeoutTimer);

                    if (wsTestCase === 'rtt')
                    {
                        measurementStart();
                        break;
                    }
    
                    if (wsTestCase === 'download')
                    {
                        wsStartupStartTime = performance.now()+500;
                        wsStartupTimeout = setTimeout(measurementStart, wsStartupTime+500);
                        break;
                    }
                    else
                    {
                        wsStartupStartTime = performance.now();
                        wsStartupTimeout = setTimeout(measurementStart, wsStartupTime);
                        break;
                    }
                }
                break;
            }

            case 'close':
            {
                if (typeof wsWorkersStatus !== 'undefined') wsWorkersStatus[data.wsID] = data.wsState;
                break;
            }

            case 'report':
            {
                if (typeof wsWorkersStatus !== 'undefined') wsWorkersStatus[data.wsID] = data.wsState;

                if (wsTestCase === 'rtt')
                {
                    if (data.wsRttAvg)           wsRttAvg        = data.wsRttAvg;
                    if (data.wsRttMed)           wsRttMed        = data.wsRttMed;
                    if (data.wsRttMin)           wsRttMin        = data.wsRttMin;
                    if (data.wsRttMax)           wsRttMax        = data.wsRttMax;
                    if (data.wsRttRequests)      wsRttRequests   = data.wsRttRequests;
                    if (data.wsRttReplies)       wsRttReplies    = data.wsRttReplies;
                    if (data.wsRttErrors)        wsRttErrors     = data.wsRttErrors;
                    if (data.wsRttMissing)       wsRttMissing    = data.wsRttMissing;
                    if (data.wsRttPacketsize)    wsRttPacketsize = data.wsRttPacketsize;

                    break;
                }

                if (data.msg === 'startupCompleted')
                {
                    if (wsTestCase === 'download')
                    {
                        if (wsFrameSize >= 65536)
                        {
                            wsOverheadPerFrame          = 8;
                            dlWsOverheadPerFrame        = wsOverheadPerFrame;
                            wsDownloadOverheadPerFrame  = wsOverheadPerFrame;
                        }
                        else if (wsFrameSize < 126)
                        {
                            wsOverheadPerFrame          = 2;
                            dlWsOverheadPerFrame        = wsOverheadPerFrame;
                            wsDownloadOverheadPerFrame  = wsOverheadPerFrame;
                        }
                        
                        dlStartupData   += data.wsData;
                        dlStartupFrames += data.wsFrames;
                    }

                    if (wsTestCase === 'upload')
                    { 
                        ulStartupData           += data.wsData;
                        ulStartupFrames         += data.wsFrames;
                    }

                    break;
                }
                else if (wsTestCase === 'download')
                {
                    dlData          += data.wsData;
                    dlFrames        += data.wsFrames;
                    
                    if (data.wsTime > wsWorkerTime) 
                    {
                        wsWorkerTime = data.wsTime;
                    }

                    break;
                }
                else if (wsTestCase === 'upload')
                {
                    var ulStreamReportDict = ulReportDict[data.wsID];

                    for (var key in data.ulReportDict)
                    {
                        var ulValueDict = {};
                        ulValueDict.bRcv = Number(data.ulReportDict[key].bRcv);
                        ulValueDict.hRcv = Number(data.ulReportDict[key].hRcv);
                        ulStreamReportDict[data.ulReportDict[key].time] = ulValueDict;
                    }
                    ulReportDict[data.wsID] = ulStreamReportDict;

                    break;
                }

                break;
            }

            case 'error':
            {
                wsWorkersStatus[data.wsID] = data.wsState;                   
                break;
            } 

            default:
            {
                console.log('wsWorker ' + data.wsID + ' unknown command: ' + data.cmd + '\' message: \'' + data.msg);
                break;
            }
        }
    }
    
    /**
     * @function measurementError
     * @description Function to report errors and close active connections
     * @private
     * @param {string} errorDescription Description of the Error
     * @param {string} errorCode Error Code
     * @param {string} systemAvailability System Availability
     * @param {string} serviceAvailability Service Availability
     */
    function measurementError(errorDescription, errorCode, systemAvailability, serviceAvailability)
    {
        clearInterval(wsTimeoutTimer);
        console.log(wsTestCase + ': ' + errorDescription);
        wsErrorCode             = errorCode;
        wsErrorDescription      = errorDescription;
        reportToMeasurement('error',  errorDescription);
        for (var wsID = 0; wsID < wsWorkers.length; wsID++)
        {
            var workerData = prepareWorkerData('close', wsID);
            
            if (!singleThread)
            {
                wsWorkers[wsID].postMessage(workerData);
            }
            else
            {
                wsWorkers[wsID].onmessage(workerData);
            }
        }
        resetValues();
    }

    /**
     * @function measurementTimeout
     * @description Function to report timeouts
     * @private
     */
    function measurementTimeout()
    {
        wsMeasurementError = true;
        measurementError('webSocket timeout error', 2, 1, 0);
    }
    
    /**
     * @function measurementStart
     * @description Function to start measurements
     * @private
     */
    function measurementStart()
    {
        wsStartTime = performance.now();
        
        if (wsTestCase !== 'rtt')
        {
            for (var wsID = 0; wsID < wsWorkers.length; wsID++)
            {
                if (wsWorkersStatus[wsID] === wsStateOpen) wsStreamsStart++;
                var workerData = prepareWorkerData('resetCounter', wsID);
                
                if (!singleThread)
                {
                    wsWorkers[wsID].postMessage(workerData); 
                }
                else
                {
                    wsWorkers[wsID].onmessage(workerData);
                }
            }
        }

        wsInterval = setInterval(measurementReport, wsReportInterval);

        console.log(wsTestCase + ': measurement started');
        reportToMeasurement('info', 'measurement started');
    }
    
    /**
     * @function measurementReport
     * @description Function to report measurement results 
     * @private
     */
    function measurementReport()
    {
        wsEndTime = performance.now();
        if (((wsEndTime - wsStartTime) > wsMeasurementRunningTime) || (wsRttReplies + wsRttMissing + wsRttErrors) === rttRequests)
        {
            clearInterval(wsInterval);
            wsCompleted = true;
            for (var wsID = 0; wsID < wsWorkers.length; wsID++)
            {
                if (wsWorkersStatus[wsID] === wsStateOpen) wsStreamsEnd++;
                var workerData = prepareWorkerData('close', wsID);
                
                if (!singleThread)
                {
                    wsWorkers[wsID].postMessage(workerData);
                }
                else
                {
                    wsWorkers[wsID].onmessage(workerData);
                }
            }
            wsEndTime = performance.now();
            setTimeout(measurementFinish, 100);
        }
        else if ((wsEndTime - wsStartTime) > 500)
        {
            for (var wsID = 0; wsID < wsWorkers.length; wsID++)
            {
                var workerData = prepareWorkerData('report', wsID);
                
                if (!singleThread)
                {                
                    wsWorkers[wsID].postMessage(workerData);
                }
                else
                {
                    wsWorkers[wsID].onmessage(workerData);
                }
            }
            wsMeasurementTime       = performance.now() - wsStartTime;
            wsMeasurementTimeTotal  = performance.now() - wsStartupStartTime;
            
            if (!wsCompleted)
            {
                setTimeout(report, 100);
            }
        }
    }
    
    /**
     * @function measurementFinish
     * @description Function to finish measurements
     * @private
     */
    function measurementFinish()
    {
        console.log(wsTestCase + ': measurement finished');

        wsMeasurementTime       = wsEndTime - wsStartTime;
        wsMeasurementTimeTotal  = wsEndTime - wsStartupStartTime;
        report(true);
    }




    /*-------------------------helper------------------------*/

    /**
     * @function report
     * @description Function to report measurement results
     * @private
     * @param {bool} finish Indicates if the measurement is finished
     */
    function report(finish)
    {
        if ((wsWorkerTime - wsStartTime) > wsMeasurementTime)
        {
            wsMeasurementTime       = wsWorkerTime - wsStartTime;
            wsMeasurementTimeTotal  = wsWorkerTime - wsStartupStartTime;
        }
            
        if (wsTestCase === 'download')
        {
            wsData          = dlData;
            wsFrames        = dlFrames;
            wsDataTotal     = dlData    + dlStartupData;
            wsFramesTotal   = dlFrames  + dlStartupFrames;       
        }
        
        var msg;
        
        if (wsTestCase === 'upload')
        {
            var ulData = 0;
            var ulFrames = 0;
            for (var wsID = 0; wsID < wsWorkers.length; wsID++)
            {
                var ulStreamReportDict = ulReportDict[wsID];
                
                for (var streamKey in ulStreamReportDict)
                {
                    
                    ulData      += ulStreamReportDict[streamKey].bRcv;
                    ulFrames    += ulStreamReportDict[streamKey].hRcv;
                }
            }
            
            wsData          = ulData;
            wsFrames        = ulFrames;
            wsDataTotal     = ulData    + ulStartupData;
            wsFramesTotal   = ulFrames  + ulStartupFrames;  
            
            msg = 'ok';
        }
        
        if (wsTestCase !== 'rtt')
        {    
            wsOverhead           = (wsFrames * wsOverheadPerFrame);
            wsOverheadTotal      = (wsFramesTotal * wsOverheadPerFrame);
            wsSpeedAvgBitS       = (((wsData * 8) + (wsOverhead * 8)) / (Math.round(wsMeasurementTime) / 1000));
            wsSpeedAvgMBitS      = +(wsSpeedAvgBitS / 1000 / 1000).toFixed(2);

            //set min/max rates
            if (wsSpeedAvgBitS < wsSpeedMinBitS)
                wsSpeedMinBitS = wsSpeedAvgBitS;
            else if (wsSpeedMinBitS === 0)
                wsSpeedMinBitS = wsSpeedAvgBitS;
            if (wsSpeedAvgBitS > wsSpeedMaxBitS)
                wsSpeedMaxBitS = wsSpeedAvgBitS;
        }

        var finishString     = '';
        var cmd              = 'report';
        
        if (finish)
        {
            cmd = 'finish';
            msg = 'measurement finished';
            finishString = 'Overall ';
            console.log('--------------------------------------------------------');
        }

        if (wsTestCase === 'rtt')
        {
            console.log(finishString + 'Time:            ' + Math.round(wsMeasurementTime) + ' ms');
            console.log(finishString + 'RTT Avg:         ' + wsRttAvg + ' ms');
            console.log(finishString + 'RTT Median:      ' + wsRttMed + ' ms');
            console.log(finishString + 'RTT Min:         ' + wsRttMin + ' ms');
            console.log(finishString + 'RTT Max:         ' + wsRttMax + ' ms');
            console.log(finishString + 'RTT Requests:    ' + wsRttRequests);
            console.log(finishString + 'RTT Replies:     ' + wsRttReplies);
            console.log(finishString + 'RTT Errors:      ' + wsRttErrors);
            console.log(finishString + 'RTT Missing:     ' + wsRttMissing);
            console.log(finishString + 'RTT Packetsize:  ' + wsRttPacketsize);
        }
        else
        {
            console.log(finishString + 'Time:            ' + Math.round(wsMeasurementTime) + ' ms');
            console.log(finishString + 'Data:            ' + (wsData + wsOverhead) + ' bytes');
            console.log(finishString + 'TCP Throughput:  ' + wsSpeedAvgMBitS + ' Mbit/s');
        }
        
        //set KPIs 
        if (wsTestCase === 'download')
        {
            wsDownloadRateAvg       = Math.round(wsSpeedAvgBitS);
            wsDownloadRateMin       = Math.round(wsSpeedMinBitS);
            wsDownloadRateMax       = Math.round(wsSpeedMaxBitS);
            wsDownloadData          = wsData + wsOverhead;
            wsDownloadDataTotal     = wsDataTotal + wsOverheadTotal;  
            wsDownloadDuration      = Math.round(wsMeasurementTime);
            wsDownloadDurationTotal = Math.round(wsMeasurementTimeTotal);
            wsDownloadStreamsStart  = wsStreamsStart;
            wsDownloadStreamsEnd    = wsStreamsEnd;
            wsDownloadFrameSize     = wsFrameSize;
            wsDownloadFrames        = wsFrames;
            wsDownloadFramesTotal   = wsFramesTotal;
            wsDownloadOverhead      = wsOverhead;
            wsDownloadOverheadTotal = wsOverheadTotal;     
            wsDownloadRateAvgMBitS  = wsSpeedAvgMBitS;
        }
        
        if (wsTestCase === 'upload')
        {
            wsUploadRateAvg         = Math.round(wsSpeedAvgBitS);
            wsUploadRateMin         = Math.round(wsSpeedMinBitS);
            wsUploadRateMax         = Math.round(wsSpeedMaxBitS);
            wsUploadData            = wsData + wsOverhead;
            wsUploadDataTotal       = wsDataTotal + wsOverheadTotal;  
            wsUploadDuration        = Math.round(wsMeasurementTime);
            wsUploadDurationTotal   = Math.round(wsMeasurementTimeTotal);
            wsUploadStreamsStart    = wsStreamsStart;
            wsUploadStreamsEnd      = wsStreamsEnd;
            wsUploadFrameSize       = wsFrameSize;
            wsUploadFrames          = wsFrames;
            wsUploadFramesTotal     = wsFramesTotal;
            wsUploadOverhead        = wsOverhead;
            wsUploadOverheadTotal   = wsOverheadTotal;  
            wsUploadRateAvgMBitS    = wsSpeedAvgMBitS;
        }
        
        reportToMeasurement(cmd, msg);
        if (finish) resetValues();
    }
    
    /**
     * @function reportToMeasurement
     * @description Function to report measurement results to the WSControl Callback class
     * @private
     * @param {string} cmd Callback Command
     * @param {string} msg Callback Message
     */
    function reportToMeasurement(cmd, msg)
    {
        var report          = {};
        report.cmd          = cmd;
        report.msg          = msg;
        report.test_case    = wsTestCase;

        if (wsTestCase === 'rtt')       report = getKPIsRtt(report);
        if (wsTestCase === 'download')  report = getKPIsDownload(report);
        if (wsTestCase === 'upload')    report = getKPIsUpload(report);
        report = getKPIsAvailability(report);

        this.wsMeasurement.controlCallback(JSON.stringify(report));
    }
    
    /**
     * @function getKPIsRtt
     * @description Function to collect RTT KPIs
     * @private
     * @param {string} report Object to add KPIs to
     */
    function getKPIsRtt(report)
    {
        report.rtt_avg                      = wsRttAvg;
        report.rtt_med                      = wsRttMed;
        report.rtt_min                      = wsRttMin;
        report.rtt_max                      = wsRttMax;
        report.rtt_requests                 = wsRttRequests;
        report.rtt_replies                  = wsRttReplies;
        report.rtt_errors                   = wsRttErrors;
        report.rtt_missing                  = wsRttMissing;
        report.rtt_packetsize               = wsRttPacketsize;
        
        return report;
    }
    
    /**
     * @function getKPIsDownload
     * @description Function to collect Download KPIs
     * @private
     * @param {string} report Object to add KPIs to
     */
    function getKPIsDownload(report)
    {
        report.download_rate_min            = wsDownloadRateMin;
        report.download_rate_avg            = wsDownloadRateAvg;
        report.download_rate_max            = wsDownloadRateMax;
        report.download_data                = wsDownloadData;
        report.download_data_total          = wsDownloadDataTotal;
        report.download_duration            = wsDownloadDuration;
        report.download_duration_total      = wsDownloadDurationTotal;
        report.download_streams_start       = wsDownloadStreamsStart;
        report.download_streams_end         = wsDownloadStreamsEnd;
        
        report.download_frame_size          = wsDownloadFrameSize;
        report.download_frames              = wsDownloadFrames;
        report.download_frames_total        = wsDownloadFramesTotal;
        report.download_overhead            = wsDownloadOverhead;
        report.download_overhead_total      = wsDownloadOverheadTotal;
        report.download_overhead_per_frame  = wsDownloadOverheadPerFrame;
        report.download_rate_avg_mbits      = wsDownloadRateAvgMBitS;

        return report;
    }
    
    /**
     * @function getKPIsUpload
     * @description Function to collect Upload KPIs
     * @private
     * @param {string} report Object to add KPIs to
     */
    function getKPIsUpload(report)
    {
        report.upload_rate_min              = wsUploadRateMin;
        report.upload_rate_avg              = wsUploadRateAvg;
        report.upload_rate_max              = wsUploadRateMax;
        report.upload_data                  = wsUploadData;
        report.upload_data_total            = wsUploadDataTotal;
        report.upload_duration              = wsUploadDuration;
        report.upload_duration_total        = wsUploadDurationTotal;
        report.upload_streams_start         = wsUploadStreamsStart;
        report.upload_streams_end           = wsUploadStreamsEnd;
        
        report.upload_frame_size            = wsUploadFrameSize;
        report.upload_frames                = wsUploadFrames;
        report.upload_frames_total          = wsUploadFramesTotal;
        report.upload_overhead              = wsUploadOverhead;
        report.upload_overhead_total        = wsUploadOverheadTotal;
        report.upload_overhead_per_frame    = wsUploadOverheadPerFrame;
        report.upload_rate_avg_mbits        = wsUploadRateAvgMBitS; 
        
        return report;
    }
    
    /**
     * @function getKPIsAvailability
     * @description Function to collect Availability KPIs
     * @private
     * @param {string} report Object to add KPIs to
     */
    function getKPIsAvailability(report)
    {
        report.error_code                   = wsErrorCode;
        report.error_description            = wsErrorDescription;
        
        return report;
    }
    
    /**
     * @function prepareWorkerData
     * @description Function to prepare the WSWorker control data
     * @private
     * @param {string} cmd Command to execute
     * @param {int} wsID ID of the WSWorker
     */
    function prepareWorkerData(cmd, wsID)
    {
        var workerData                  = {};
        workerData.cmd                  = cmd;
        workerData.wsTestCase           = wsTestCase;
        workerData.wsID                 = wsID;
        workerData.wsTarget             = wsTarget;
        workerData.wsRttTarget          = wsRttTarget;
        workerData.wsTargetPort         = wsTargetPort;
        workerData.wsProtocol           = wsProtocol;
        workerData.wsFrameSize          = wsFrameSize;
        
        if (wsTestCase === 'rtt')
        {
            workerData.rttRequests       = rttRequests;
            workerData.rttRequestTimeout = rttRequestTimeout;
            workerData.rttRequestWait    = rttRequestWait;
            workerData.rttTimeout        = rttTimeout;
            workerData.rttPayloadSize    = rttPayloadSize;
        }    
        
        return JSON.stringify(workerData);
    }
    
    /**
     * @function resetValues
     * @description Initialize all variables with default values
     * @private
     */
    function resetValues()
    {
        clearInterval(wsInterval);

        wsTestCase                  = '';
        wsData                      = 0;
        wsDataTotal                 = 0;
        wsFrames                    = 0;
        wsFramesTotal               = 0;
        wsSpeedAvgBitS              = 0;
        wsSpeedAvgMBitS             = 0;
        wsSpeedMinBitS              = 0;
        wsSpeedMaxBitS              = 0;
        wsOverhead                  = 0;
        wsOverheadTotal             = 0;
        wsStartTime                 = performance.now();
        wsEndTime                   = performance.now();
        wsMeasurementTime           = 0;
        wsMeasurementTimeTotal      = 0;
        wsStartupStartTime          = performance.now();
        wsOverheadPerFrame          = 0;
        wsWorkers                   = 0;
        wsWorkersStatus             = 0;
        wsWorkerTime                = 0;
        wsInterval                  = 0;
        wsMeasurementRunningTime    = 0;
        wsCompleted                 = false;
        wsReportInterval            = 0;
        wsTimeoutTimer              = 0;
        wsTimeout                   = 0;
        wsStartupTimeout            = 0;
        wsStreamsStart              = 0;
        wsStreamsEnd                = 0;
        wsFrameSize                 = 0;
        
        dlData                      = 0;
        dlFrames                    = 0;
        dlStartupData               = 0;
        dlStartupFrames             = 0;
        
        ulReportDict                = {};
        ulStartupData               = 0;
        ulStartupFrames             = 0;
        
        
        
        
        /*-------------------------KPIs------------------------*/
        
        wsRttAvg                    = 0;
        wsRttMed                    = 0;
        wsRttMin                    = 0;
        wsRttMax                    = 0;
        wsRttRequests               = 0;
        wsRttReplies                = 0;
        wsRttErrors                 = 0;
        wsRttMissing                = 0;
        wsRttPacketsize             = 0;

        wsDownloadRateMin           = 0;
        wsDownloadRateAvg           = 0;
        wsDownloadRateMax           = 0;
        wsDownloadData              = 0;
        wsDownloadDataTotal         = 0;
        wsDownloadDuration          = 0;
        wsDownloadDurationTotal     = 0;
        wsDownloadStreamsStart      = 0;
        wsDownloadStreamsEnd        = 0;
        
        wsDownloadFrameSize         = 0;
        wsDownloadFrames            = 0;
        wsDownloadFramesTotal       = 0;
        wsDownloadOverheadPerFrame  = dlWsOverheadPerFrame;
        wsDownloadOverhead          = 0;
        wsDownloadOverheadTotal     = 0;
        wsDownloadRateAvgMBitS      = 0;
        
        wsUploadRateMin             = 0;
        wsUploadRateAvg             = 0;
        wsUploadRateMax             = 0;
        wsUploadData                = 0;
        wsUploadDataTotal           = 0;
        wsUploadDuration            = 0;
        wsUploadDurationTotal       = 0;
        wsUploadStreamsStart        = 0;
        wsUploadStreamsEnd          = 0;
        
        wsUploadFrameSize           = 0;
        wsUploadFrames              = 0;
        wsUploadFramesTotal         = 0;
        wsUploadOverheadPerFrame    = ulWsOverheadPerFrame;
        wsUploadOverhead            = 0;
        wsUploadOverheadTotal       = 0;
        wsUploadRateAvgMBitS        = 0;

        wsErrorCode                 = 0;
        wsErrorDescription          = '-';
    }
};