/*!   
    \file WSMeasurement.js
    \brief WebSocket Measurement Class
    \author zafaco GmbH <info@zafaco.de>
    \date Last update: 2017-11-24
    \version 1.1

    Copyright (C) 2016 - 2017 zafaco GmbH

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3 
    as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

var wsControl;
var jsTool = new JSTool();




/*-------------------------class WSMeasurement------------------------*/

/**
 * @class WSMeasurement
 * @description WebSocket Measurement Class
 */
function WSMeasurement()
{
    //Test Cases
    var performRttMeasurement               = false;
    var performDownloadMeasurement          = false;
    var performUploadMeasurement            = false;
    
    var performedRttMeasurement             = false;
    var performedDownloadMeasurement        = false;
    var performedUploadMeasurement          = false;
    
    var wsMeasurementParameters             = {};
    
    //KPIs
    var availabilityKPIs                    = {};
    var rttKPIs                             = {};
    var downloadKPIs                        = {};
    var uploadKPIs                          = {};
    
    
    
    
    /*-------------------------public functions------------------------*/
    
    /**
     * @function measurementControl
     * @description API Function to Start and Stop measurements
     * @public
     * @param {string} measurementParameters JSON coded measurement Parameters
     */
    this.measurementControl = function(measurementParameters)
    {
        wsMeasurementParameters = JSON.parse(measurementParameters);

        resetWsControl();

        if ((wsMeasurementParameters.cmd) === 'start')
        {
            var userAgent = navigator.userAgent;
            //catch Firefox < 38, as it has no WebSocket in WebWorker Support
            if (((userAgent.search('Firefox') !== -1) || (userAgent.search('Seamonkey') !== -1)) && (Number(userAgent.match(/Firefox\/((\d+\.)+\d+)/)[1]) < 38))
            {
                wsMeasurementParameters.singleThread    = true;
            }
            
            //catch IE 11 and Edge, as they do not have WebSocket in WebWorker Support
            if ((userAgent.search('Edge') !== -1) || (userAgent.search('MSIE') !== -1) || (userAgent.search('Trident') !== -1))
            {
                wsMeasurementParameters.singleThread    = true;
            }
        }

        if (typeof wsMeasurementParameters.performRttMeasurement !== 'undefined')           performRttMeasurement           = Boolean(wsMeasurementParameters.performRttMeasurement);
        if (typeof wsMeasurementParameters.performDownloadMeasurement !== 'undefined')      performDownloadMeasurement      = Boolean(wsMeasurementParameters.performDownloadMeasurement);
        if (typeof wsMeasurementParameters.performUploadMeasurement !== 'undefined')        performUploadMeasurement        = Boolean(wsMeasurementParameters.performUploadMeasurement);
        
        if (!window.WebSocket)
        {
            var data = {};
            data.cmd                                = 'error';
            availabilityKPIs.cmd                    = 'error';     
            availabilityKPIs.error_code             = 3;
            availabilityKPIs.error_description      = 'WebSockets are not supported by your browser';
            data                                    = JSON.stringify(data);
            this.controlCallback(data);
            return;
        }
        
        switch (wsMeasurementParameters.cmd)
        {
            case 'start':
            {
                initializeKPIs();
                measurementCampaign();
                
                break;
            }
        }
    };
    
    /**
     * @function controlCallback
     * @description API Function to Callback the JSON coded measurement Results to the implementing client
     * @param {string} data JSON coded measurement Results
     */
    this.controlCallback = function(data)
    {
        data = JSON.parse(data);

        if (data.test_case === 'rtt')           rttKPIs             = data;
        if (data.test_case === 'download')      downloadKPIs        = data;
        if (data.test_case === 'upload')        uploadKPIs          = data;

        if (data.cmd === 'error')
        {
            setTimeout(resetWsControl, 200);
        }

        if (data.cmd === 'finish')
        {
            if (data.test_case === 'rtt')           
            {
                performedRttMeasurement     = true;
            }

            if (data.test_case === 'download')
            {
                performedDownloadMeasurement        = true;
            }
            
            if (data.test_case === 'upload')
            {
                performedUploadMeasurement          = true;
            }
            
            measurementCampaign();
        }
        
        var kpis = getKPIs();
        
        if (performRttMeasurement === performedRttMeasurement && performDownloadMeasurement === performedDownloadMeasurement && performUploadMeasurement === performedUploadMeasurement)
        {
            var kpisCompleted = jsTool.extend(kpis);
            kpisCompleted.cmd = 'completed';
            kpisCompleted = JSON.stringify(kpisCompleted);
            setTimeout(reportToWeb, 50, kpisCompleted);
        }
         
        kpis = JSON.stringify(kpis);
        
        reportToWeb(kpis);
    };
    
    
    
    
    /*-------------------------private functions------------------------*/
    
    /**
     * @function measurementCampaign
     * @description Perform a measurement Campaign containing one or more Test Cases
     * @private
     */
    function measurementCampaign()
    {
        wsControl               = null;
        delete wsControl;
        if (typeof wsMeasurementParameters.singleThread !== 'undefined')
        {
            wsControl               = new WSControlSingleThread();
        }
        else
        {
            wsControl               = new WSControl();
        }
        wsControl.wsMeasurement = this;
        
        var waitTime            = 3000;
        var waitTimeShort       = 1000;
        
        if (performRttMeasurement && !performedRttMeasurement)
        {
            wsMeasurementParameters.testCase = 'rtt';
            setTimeout(wsControl.measurementSetup, waitTimeShort, JSON.stringify(wsMeasurementParameters));

            return;
        }

        if (performDownloadMeasurement && !performedDownloadMeasurement)
        { 
            wsMeasurementParameters.testCase = 'download';
            setTimeout(wsControl.measurementSetup, waitTimeShort, JSON.stringify(wsMeasurementParameters));
            
            return;
        }
        else
        if (performUploadMeasurement && !performedUploadMeasurement)
        {
            wsMeasurementParameters.testCase = 'upload';
            setTimeout(wsControl.measurementSetup, waitTime, JSON.stringify(wsMeasurementParameters));
           
            return;
        }  
        else
        {
            setTimeout(resetWsControl, 200);
        }
    }

    /**
     * @function initializeKPIs
     * @description Initialize all KPIs with default values
     * @private
     */
    function initializeKPIs()
    {
        availabilityKPIs.error_code                         = 0;
        availabilityKPIs.error_description                  = '-';
         
        rttKPIs.rtt_avg                                     = 0;
        rttKPIs.rtt_med                                     = 0;
        rttKPIs.rtt_min                                     = 0;
        rttKPIs.rtt_max                                     = 0;
        rttKPIs.rtt_requests                                = 0;
        rttKPIs.rtt_replies                                 = 0;
        rttKPIs.rtt_errors                                  = 0;
        rttKPIs.rtt_missing                                 = 0;
        rttKPIs.rtt_packetsize                              = 0;
        
        downloadKPIs.download_rate_min                      = 0;
        downloadKPIs.download_rate_avg                      = 0;
        downloadKPIs.download_rate_max                      = 0;
        downloadKPIs.download_data                          = 0;
        downloadKPIs.download_data_total                    = 0;
        downloadKPIs.download_duration                      = 0;
        downloadKPIs.download_duration_total                = 0;
        downloadKPIs.download_streams_start                 = 0;
        downloadKPIs.download_streams_end                   = 0;
        downloadKPIs.download_frame_size                    = 0;
        downloadKPIs.download_frames                        = 0;
        downloadKPIs.download_frames_total                  = 0;
        downloadKPIs.download_overhead                      = 0;
        downloadKPIs.download_overhead_total                = 0;
        downloadKPIs.download_overhead_per_frame            = 0;
        downloadKPIs.download_rate_avg_mbits                = 0;
        
        uploadKPIs.upload_rate_min                          = 0;
        uploadKPIs.upload_rate_avg                          = 0;
        uploadKPIs.upload_rate_max                          = 0;
        uploadKPIs.upload_data                              = 0;
        uploadKPIs.upload_data_total                        = 0;
        uploadKPIs.upload_duration                          = 0;
        uploadKPIs.upload_duration_total                    = 0;
        uploadKPIs.upload_streams_start                     = 0;
        uploadKPIs.upload_streams_end                       = 0;
        uploadKPIs.upload_frame_size                        = 0;
        uploadKPIs.upload_frames                            = 0;
        uploadKPIs.upload_frames_total                      = 0;
        uploadKPIs.upload_overhead                          = 0;
        uploadKPIs.upload_overhead_total                    = 0;
        uploadKPIs.upload_overhead_per_frame                = 0;
        uploadKPIs.upload_rate_avg_mbits                    = 0;
    }
   
    /**
     * @function getKPIs
     * @description Return all measurement Results
     */
    function getKPIs()
    {
        var kpis = jsTool.extend(availabilityKPIs, rttKPIs, downloadKPIs, uploadKPIs);

        return kpis;
    }

    /**
     * @function resetWsControl
     * @description Reset the wsControl object
     * @private
     */
    function resetWsControl()
    {
        wsControl = null;
        delete wsControl;
    }

    /**
     * @function reportToWeb
     * @description Callback to Browser
     * @private
     * @param {string} kpis JSON coded Measurement Results
     */
    function reportToWeb(kpis)
    {
        measurementCallback(kpis);
    }
};