/*!   
    \file WSWorker.js
    \brief WebSocket Worker Class
    \author zafaco GmbH <info@zafaco.de>
    \date Last update: 2017-11-24

    Copyright (C) 2016 - 2017 zafaco GmbH

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3 
    as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/




/*-------------------------class WSWorker------------------------*/

/**
 * @class WSWorker
 * @description WebSocket Worker Class
 */
var WSWorker = function ()
{
    this.wsControl      = 'undefined';

    return(this);
};

var webSocket;

var wsTarget;
var wsRttTarget;
var wsTargetPort;
var wsProtocol;
var wsTestCase;
var wsID;
var wsFrameSize;
var wsData;
var wsFrames;
var wsCompleted;
var wsServer;

var wsStateConnecting   = 0;
var wsStateOpen         = 1;
var wsStateClosing      = 2;
var wsStateClosed       = 3;

var wsConnected         = false;

var wsRttAvg;
var wsRttMed;
var wsRttMin;
var wsRttMax;
var wsRttRequests;
var wsRttReplies;
var wsRttErrors;
var wsRttMissing;
var wsRttPacketsize;

var rttRequests             = 10;
var rttRequestTimeout       = 1000;
var rttRequestWait          = 500;
var rttTimeout              = ((rttRequests * rttRequestTimeout) + rttRequestWait) * 1.3;
var rttPayloadSize          = 64;

var ulData;
var ulDataPointer;
var ulInterval;
var ulReportDict            = {};

var ulDataSize              = 2246915;
var ulBufferSize            = 4096 * 1000;
var ulStarted               = false;

var singleThread            = false;

var jsTool;




/*-------------------------WSWorker------------------------*/

/**
 * @function onmessage
 * @description Function to Receive Commands from WSControl Single Thread 
 * @public
 * @param {string} data JSON coded measurement Parameters
 */
WSWorker.prototype.onmessage = function(data)
{
    singleThread = true;
    onmessage(data);
};

/**
 * @function onmessage
 * @description Function to Receive Commands from WSControl Multi Thread 
 * @public
 * @param {string} event JSON coded measurement Parameters
 */
onmessage = function (event)
{
    var data;

    try 
    {
        if (!singleThread) 
        {
            data = JSON.parse(event.data);
        }
        else
        {
            data = JSON.parse(event);
            wsTestCase = data.wsTestCase;
        }
    }
    catch (error)
    {
        return;
    }

    wsID    = data.wsID;

    switch (data.cmd)
    {
        case 'connect':
        {
            reportToControl('info', 'websocket connecting');
            setWsParameters(data);

            if (wsTestCase === 'upload') wsFrameSize = data.wsFrameSize;
            
            if (wsTestCase === 'rtt')
            {
                rttRequests       = data.rttRequests;
                rttRequestTimeout = data.rttRequestTimeout;
                rttRequestWait    = data.rttRequestWait;
                rttTimeout        = data.rttTimeout;
                rttPayloadSize    = data.rttPayloadSize;
            }

            connect();
            break;
        }

        case 'resetCounter':
        {
            if (wsTestCase === 'upload')
            {
                for (var key in ulReportDict)
                {
                    wsData      += ulReportDict[key].bRcv;
                    wsFrames    += ulReportDict[key].hRcv;
                }
            }

            reportToControl('report', 'startupCompleted');
            resetCounter();
            reportToControl('info', 'counter reseted');
            break;
        }

        case 'report':
        {
            reportToControl('report');
            resetCounter();
            break;
        }

        case 'stop':
        {
            break;
        }

        case 'close':
        {   
            if (wsTestCase !== 'loadbalancer') reportToControl('report');

            wsCompleted = true;

            if (webSocket.readyState === wsStateClosed)
            {
                webSocket.close();
                if (!singleThread) this.close();
            }
            else
            {
                webSocket.close();
                websocketClose('close received');
                if (wsTestCase !== 'loadbalancer') reportToControl('info', 'websocket closing');   
            }
            break;
        }

        default:
        {
            reportToControl('error', 'Unknown command: ' + data.msg);
            break;
        }
    }
};

/**
 * @function connect
 * @description Function to initiate a WebSocket Connection
 * @private
 */
function connect()
{   
    resetValues();

    if (!singleThread) importScripts('libs/JSTool.js');
    jsTool = new JSTool();

    var wsWssString                 = 'ws://';

    var target = wsWssString + wsTarget + ':' + wsTargetPort;    

    if (wsTestCase === 'rtt')
    {
        target = wsWssString + wsRttTarget + ':' + wsTargetPort;
    }

    try 
    {
        webSocket = new WebSocket(target, wsProtocol);
    }
    catch (error)
    {
        console.log('webSocket error: try/catch: ' + error);
    }

    webSocket.onopen = function (event)
    {
        reportToControl('open', 'websocket open');
        wsConnected = true;

        if (wsTestCase === 'rtt')
        {
            roundTripTime();
        }
        else if (wsTestCase === 'download')
        {
            download();
        } 
        else if (wsTestCase === 'upload')
        {
            setTimeout(upload, 200);
            wsCompleted = false;
            ulData = jsTool.generateRandomData(ulDataSize, true);
        }
    };

    webSocket.onerror = function (event)
    {
        console.log('webSocket error: onError: ' + event.type);
    };

    webSocket.onmessage = function (event)
    {  
        if (wsTestCase === 'download')
        {
            wsData += event.data.size;
            wsFrames++;
        }
        else
        {
            try 
            {
                var data = JSON.parse(event.data);
            }
            catch (error)
            {
                return;
            }

            if (data.cmd === 'rttReport')
            {
                wsRttAvg        = Number(data.avg);
                wsRttMed        = Number(data.med);
                wsRttMin        = Number(data.min);
                wsRttMax        = Number(data.max);
                wsRttRequests   = Number(data.req);
                wsRttReplies    = Number(data.rep);
                wsRttErrors     = Number(data.err);
                wsRttMissing    = Number(data.mis);
                wsRttPacketsize = Number(data.pSz);
            }

            if (data.cmd === 'ulReport')
            {
                var length = 0;
                for (var key in ulReportDict)
                {
                    length++;
                }
                ulReportDict[length]        = {};
                ulReportDict[length].time   = Number(data.time);
                ulReportDict[length].bRcv   = Number(data.bRcv);
                ulReportDict[length].hRcv   = Number(data.hRcv);
            }
        }
    };

    webSocket.onclose = function (event)
    {
        wsTestCase = '';
        reportToControl('close', 'websocket closed, reason: ' + jsTool.webSocketCloseReasons(event));
    };
};

/**
 * @function roundTripTime
 * @description Function to perform the Round-Trip Time Test Case
 * @private
 */
function roundTripTime()
{
    sendToWebSocket('rttStart');
    reportToControl('info', 'start round trip time');
};

/**
 * @function download
 * @description Function to perform the Download Test Case
 * @private
 */
function download()
{
    reportToControl('info', 'start download');
};

/**
 * @function upload
 * @description Function to perform the Upload Test Case
 * @private
 */
function upload()
{
    if (!ulStarted)
    {
        wsCompleted = false;
        reportToControl('info', 'start upload'); 
        ulStarted = true;
    }

    ulInterval = setInterval(function ()
    {
        if (webSocket.bufferedAmount <= ulBufferSize)
        {
            if (wsCompleted)
            {
                clearInterval(ulInterval);
            }
                
            var ulPayload = ulData.slice(ulDataPointer, ulDataPointer + wsFrameSize);
            ulDataPointer += wsFrameSize;
            if (ulDataPointer > ulDataSize)
            {
                ulDataPointer = ulDataPointer - ulDataSize;
                ulPayload = ulPayload + ulData.slice(0, ulDataPointer);
            }
            
            if (webSocket.readyState === wsStateOpen) 
            {
                webSocket.send(ulPayload);
            }
        }
    }, 4);
};

/**
 * @function reportToControl
 * @description Function to report measurement results to the WSControl Callback class
 * @private
 * @param {string} cmd Callback Command
 * @param {string} msg Callback Message
 */
function reportToControl(cmd, msg)
{
    var data            = {};
    data.cmd            = cmd;
    data.msg            = '';
    if (msg) data.msg   = msg;
    data.wsID           = wsID;
    data.wsFrameSize    = wsFrameSize;

    if (webSocket) data.wsState    = webSocket.readyState;

    if (wsTestCase === 'rtt')
    {
        data.wsRttAvg        = wsRttAvg;
        data.wsRttMed        = wsRttMed;
        data.wsRttMin        = wsRttMin;
        data.wsRttMax        = wsRttMax;
        data.wsRttRequests   = wsRttRequests;
        data.wsRttReplies    = wsRttReplies;
        data.wsRttErrors     = wsRttErrors;
        data.wsRttMissing    = wsRttMissing;
        data.wsRttPacketsize = wsRttPacketsize;
    }

    if (wsTestCase === 'upload')
    {
        data.ulReportDict = ulReportDict;
    }

    data.wsData         = wsData;
    data.wsFrames       = wsFrames;

    if (typeof performance !== 'undefined')
    {
        data.wsTime = performance.now();
    }
    else 
    {
        data.wsTime = 0;
    }

    if (!singleThread)
    {
        self.postMessage(JSON.stringify(data));
    }
    else
    {
        wsControl.workerCallback(JSON.stringify(data));
    }
};

/**
 * @function sendToWebSocket
 * @description Function to send data to an active WebSocket connection
 * @private
 * @param {string} cmd Command
 * @param {string} msg Message
 */
function sendToWebSocket(cmd, msg)
{
    var data    = {};
    data.cmd    = cmd;
    data.msg    = msg;

    if (wsTestCase === 'rtt')
    {
        data.rttRequests       = rttRequests;
        data.rttRequestTimeout = rttRequestTimeout;
        data.rttRequestWait    = rttRequestWait;
        data.rttTimeout        = rttTimeout;
        data.rttPayloadSize    = rttPayloadSize;
    }

    if (webSocket.readyState === wsStateOpen) webSocket.send(JSON.stringify(data));
};

/**
 * @function setWsParameters
 * @description Function to set WebSocket parameters
 * @private
 * @param {string} data Parameters
 */
function setWsParameters(data)
{
    wsTarget        = data.wsTarget;
    wsRttTarget     = data.wsRttTarget;
    wsTargetPort    = data.wsTargetPort;
    wsProtocol      = data.wsProtocol;
    wsTestCase      = data.wsTestCase;
};

/**
 * @function resetCounter
 * @description Function to reset the Download/Upload Counter
 * @private
 */
function resetCounter()
{
    wsData          = 0;
    wsFrames        = 0;

    ulReportDict    = {};
};

/**
 * @function websocketClose
 * @description Function to close an active WebSocket connection
 * @private
 * @param {string} reason Close reason
 */
function websocketClose(reason)
{
    webSocket.close();
    console.log('close websocket: reason: ' + reason);
    if (!singleThread) setTimeout(this.close, 200);
};

/**
 * @function resetValues
 * @description Initialize all variables with default values
 * @private
 */
function resetValues()
{
    clearInterval(ulInterval);

    delete(webSocket);

    wsCompleted     = false;

    wsData          = 0;
    wsFrames        = 0;

    ulData          = 0;
    ulDataPointer   = 0;
    ulReportDict    = {};
};