/*!   
    \file JSTool.js
    \brief Javascript Tool Class
    \author zafaco GmbH <info@zafaco.de>
    \date Last update: 2017-07-19

    Copyright (C) 2016 - 2017 zafaco GmbH

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3 
    as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/




/*-------------------------class JSTool------------------------*/

/**
 * @class JSTool
 * @description Javascript Tool Class
 */
var JSTool = function () {
    
    return(this);
};




/*-------------------------helper------------------------*/

/**
 * @function generateRandomData
 * @description Function to generate Random ASCII Data
 * @public
 * @param {int} length Random Data Length
 * @param {bool} asString Indicates if the callback should be a string
 */
JSTool.prototype.generateRandomData = function (length, asString)
{
    var mask = '';
    mask += 'abcdefghijklmnopqrstuvwxyz';
    mask += 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
    mask += '0123456789';
    mask += '~`!@#$%^&*()_+-={}[]:\';"<>?,./|\\';
    var data = '';
    for (var i = length; i > 0; --i)
    {
        data += mask[Math.floor(Math.random() * mask.length)];
    }
    if (asString) return data;
    else return new Blob([data]);
};

/**
 * @function getTimestamp
 * @description Function to get the current Timestamp
 * @public
 */
JSTool.prototype.getTimestamp = function()
{
    if (!Date.now)
    {
        Date.now = function()
        { 
            return new Date().getTime(); 
        };
    }
    
    return Date.now();
};

/**
 * @function getFormattedDate
 * @description Function to get the current Timestamp as formatted Date
 * @public
 */
JSTool.prototype.getFormattedDate = function ()
{
    var date = new Date();
    var formattedDate;
    formattedDate = date.getFullYear() + '-' + ('0' + (date.getMonth() + 1)).slice(-2) + '-' + ('0' + date.getDate()).slice(-2) + ' ' + ('0' + date.getHours()).slice(-2) + ':' + ('0' + date.getMinutes()).slice(-2) + ':' + ('0' + date.getSeconds()).slice(-2); 
    
    return formattedDate;
};

/**
 * @function getTimezone
 * @description Function to get the current local timezone
 * @public
 */
JSTool.prototype.getTimezone = function ()
{
    var date = new Date();
    var offset = date.getTimezoneOffset();
    return offset *= -60;
};

/**
 * @function webSocketCloseReasons
 * @description Function to get the WebSocket Close Reason
 * @public
 * @param {string} event WebSocket close event
 */
JSTool.prototype.webSocketCloseReasons = function (event)
{
    switch (event.code)
    {
        case 1000: 
            return 'Normal closure, meaning that the purpose for which the connection was established has been fulfilled.';
            break;
        case 1001: 
            return 'An endpoint is \"going away\", such as a server going down or a browser having navigated away from a page.';
            break;
        case 1002: 
            return 'An endpoint is terminating the connection due to a protocol error';
            break;
        case 1003:
            return 'An endpoint is terminating the connection because it has received a type of data it cannot accept (e.g., an endpoint that understands only text data MAY send this if it receives a binary message).';
            break;
        case 1004:
            return 'Reserved. The specific meaning might be defined in the future.';
            break;
        case 1005:
            return 'No status code was actually present.';
            break;
        case 1006:
            return 'The connection was closed abnormally, e.g., without sending or receiving a Close control frame';
            break;
        case 1007:
            return 'An endpoint is terminating the connection because it has received data within a message that was not consistent with the type of the message (e.g., non-UTF-8 [http://tools.ietf.org/html/rfc3629] data within a text message).';
            break;
        case 1008:
            return  'An endpoint is terminating the connection because it has received a message that \'violates its policy\'. This reason is given either if there is no other sutible reason, or if there is a need to hide specific details about the policy.';
            break;
        case 1009:
            return 'An endpoint is terminating the connection because it has received a message that is too big for it to process.';
            break;
        case 1010:
            return 'An endpoint (client) is terminating the connection because it has expected the server to negotiate one or more extension, but the server didn\'t return them in the response message of the WebSocket handshake. <br /> Specifically, the extensions that are needed are: ' + event.reason;
            break;
        case 1011:
            return 'A server is terminating the connection because it encountered an unexpected condition that prevented it from fulfilling the request.';
            break;
        case 1015:
            return 'The connection was closed due to a failure to perform a TLS handshake (e.g., the server certificate can\'t be verified).';
            break;    
        default:
            return 'Unknown reason';
            break;
    }
};

/**
 * @function extend
 * @description Function merge two or more objects
 * @public
 * @params {object} objects to merge
 */
JSTool.prototype.extend = function ()
{
    var extended = {};
    var length = arguments.length;

    var merge = function (obj)
    {
        for (var prop in obj)
        {
            if (Object.prototype.hasOwnProperty.call(obj, prop))
            {
                extended[prop] = obj[prop];
            }
        }
    };

    for (var i=0; i < length; i++)
    {
        var obj = arguments[i];
        merge(obj);
    }

    return extended;
};