/*!   
    \file wshandler.cpp
    \brief WebSocket handler
    \author zafaco GmbH <info@zafaco.de>
    \date Last update: 2017-11-28

    Copyright (C) 2016 - 2017 zafaco GmbH

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3 
    as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "wshandler.h"


using namespace std;
using namespace json11;




int                 mSocket;
sockaddr_in6        *mClient;
        
vector<string>      allowedProtocols;
unsigned long long  webSocketTimeout;
string              sClientIp;

string              hostname;

bool                rttRunning;
bool                downloadRunning;
bool                uploadRunning;

int                 downloadFrameSize;
long long           downloadRandomDataSize;

int                 rttRequests;
unsigned long long  rttRequestTimeout;
long                rttRequestWait;
unsigned long long  rttTimeout;
int                 rttPayloadSize;
string              rttPayload;
string              rttPayloadTimestamp;
string              rttPayloadDelimiter;
bool                rttStart;
unsigned long long  rttPingSendTime;
unsigned long long  rttPongReceiveTime;
unsigned long long  rttPongAllegedReceiveTime;
vector<long long>   rttVector;

double              rttAvg;
double              rttMin;
double              rttMax;
int                 rttRequestsSend;
int                 rttReplies;
int                 rttErrors;
int                 rttMissing;
int                 rttPacketsize;
double              rttMed;
                            
long long           uploadBytesReceived;
long long           uploadHeaderReceived;

bool                showShutdown;
bool                showStopped;
bool                showRttStart;




CWsHandler::CWsHandler()
{
}


CWsHandler::~CWsHandler()
{
}


CWsHandler::CWsHandler(sockaddr_in6 *pClient, int nSocket, string nClientIp)
{
    mClient                     = pClient;
    mSocket                     = nSocket;
    
    webSocketTimeout            = 25;
    
    allowedProtocols.push_back("wsRoundTripTimeProtocol");
    allowedProtocols.push_back("wsDownloadProtocol");
    allowedProtocols.push_back("wsUploadProtocol");
    
    sClientIp                   = nClientIp;
    
    hostname                    = CTool::getHostname();
    
    rttRunning                  = false;
    downloadRunning             = false;
    uploadRunning               = false;
    
    downloadFrameSize           = 32768;
    downloadRandomDataSize      = 11234571;

    rttRequests                 = 10+1;
    rttRequestTimeout           = 1000;
    rttRequestWait              = 500;
    rttTimeout                  = (rttRequests * (rttRequestTimeout + rttRequestWait)) * 1.3;
    rttPayloadDelimiter         = ";";
    rttPayloadSize              = 64;
    rttStart                    = false;
    rttPingSendTime             = 0;
    rttPongReceiveTime          = 0;
    rttPongAllegedReceiveTime   = 0;
    rttVector.clear();

    rttAvg                      = 0;
    rttMin                      = 0;
    rttMax                      = 0;
    rttRequestsSend             = 0;
    rttReplies                  = 0;
    rttErrors                   = 0;
    rttMissing                  = 0;
    rttPacketsize               = 0;
    rttMed                      = 0;
    
    uploadBytesReceived         = 0;
    uploadHeaderReceived        = 0;
    
    showShutdown                = false;
    showStopped                 = true;
    showRttStart                = true;
}


int CWsHandler::handle_websocket()
{
    TRC_DEBUG("WebSocket handler: started");
        
    noPollCtx *ctx = nopoll_ctx_new();
    if (!ctx)
    {
        TRC_ERR("WebSocket handler: noPoll context initialization failed");
        return 1;
    }
    
    nopoll_log_enable(ctx, nopoll_true);
    nopoll_log_set_handler(ctx, nopoll_logging_handler, NULL);

    noPollConn *conn = nopoll_listener_from_socket(ctx, (NOPOLL_SOCKET)mSocket);
    
    if (!nopoll_conn_is_ok(conn))
    {
        TRC_ERR("WebSocket handler: noPoll connection initialization failed");
        return 1;
    }
    
    nopoll_ctx_set_on_open(ctx, websocket_open_handler, NULL);
    nopoll_ctx_set_on_ready(ctx, websocket_ready_handler, NULL);
    nopoll_ctx_set_on_msg(ctx, websocket_message_handler, NULL);
  
    nopoll_loop_wait(ctx, 0);

    nopoll_ctx_unref(ctx);
    nopoll_cleanup_library();
    
    if (showStopped)
    {
        TRC_DEBUG("WebSocket handler: stopped");
    }
    
    return 0;
}


int CWsHandler::websocket_open_handler(noPollCtx *ctx, noPollConn *conn, noPollPtr user_data)
{
    TRC_DEBUG("WebSocket handler: open");
    
    nopoll_conn_set_on_close(conn, websocket_close_handler, NULL);
    nopoll_conn_set_sock_block(nopoll_conn_socket(conn), nopoll_true);
    
    string sProtocol = string(nopoll_conn_get_requested_protocol(conn));
    CTool::replaceStringInPlace(sProtocol, " ", "");
    
    vector<string> requestestedProtocols;
    string delimiter = ",";
    CTool::tokenize(sProtocol, requestestedProtocols, delimiter);
    
    bool protocolAllowed = false;
    
    string acceptedProtocol;
    
    for (vector<string>::iterator itAllowedProtocols = allowedProtocols.begin(); itAllowedProtocols != allowedProtocols.end(); ++itAllowedProtocols)
    {
        for (vector<string>::iterator itRequestedProtocols = requestestedProtocols.begin(); itRequestedProtocols != requestestedProtocols.end(); ++itRequestedProtocols)
        {
            string requestedProtocol = *itRequestedProtocols;
            if (requestedProtocol.compare(*itAllowedProtocols) == 0)
            {
                acceptedProtocol    = requestestedProtocols.at(0);
                protocolAllowed     = true;

                break;
            }
        }
    }
    
    if (protocolAllowed)
    {
        nopoll_conn_set_accepted_protocol(conn, acceptedProtocol.c_str());
        TRC_DEBUG("WebSocket handler: requested protocol: \"" + acceptedProtocol + "\" is allowed");
    }
    else if (!protocolAllowed)
    {
        TRC_ERR("WebSocket handler: requested protocol: \"" + sProtocol + "\" is not allowed");

        char response[] = "HTTP/1.1 400 Bad Request\r\n\r\n";
        int responseSize = strlen(response);

        nopoll_conn_default_send(conn, response, responseSize);
        
        return 0;
    }

    return 1;
}


int CWsHandler::websocket_ready_handler(noPollCtx *ctx, noPollConn *conn, noPollPtr user_data)
{
    TRC_DEBUG("WebSocket handler: ready for IP: " + sClientIp + " on Port: " + string(nopoll_conn_port(conn)));
    
    string sProtocol = string(nopoll_conn_get_accepted_protocol(conn));

    if (sProtocol.compare("wsRoundTripTimeProtocol") == 0)
    {
        int on = 1;
        setsockopt(mSocket, IPPROTO_TCP, TCP_NODELAY, (void *) &on, sizeof (on));
        thread rttThread(roundTripTime, ctx, conn);
        rttThread.detach();
    }

    if (sProtocol.compare("wsDownloadProtocol") == 0)
    {
        thread downloadThread(download, ctx, conn);
        downloadThread.detach();
    }
    
    if (sProtocol.compare("wsUploadProtocol") == 0)
    {
        thread uploadThread(upload, ctx, conn);
        uploadThread.detach();
    }
    
    return 1;
}

void CWsHandler::websocket_message_handler(noPollCtx *ctx, noPollConn *conn, noPollMsg *msg, noPollPtr user_data)
{   
    if (nopoll_msg_opcode(msg) == NOPOLL_PONG_FRAME && rttRunning)
    {    
        rttPongAllegedReceiveTime = CTool::get_timestamp();
        
        string rttPayloadReceivedString = (const char*)nopoll_msg_get_payload(msg);
 
        size_t rttPayloadReceivedPos;
        if ((rttPayloadReceivedPos = rttPayloadReceivedString.find(rttPayloadDelimiter)) != string::npos)
        {
            TRC_DEBUG("WebSocket handler: rtt PONG received");
            
            string rttPayloadReceivedTimestamp = rttPayloadReceivedString.substr(0, rttPayloadReceivedPos);
            
            if (rttPayloadReceivedTimestamp.compare(rttPayloadTimestamp) == 0)
            {
                rttPongReceiveTime = rttPongAllegedReceiveTime;
                return;
            }
            else
            {
                TRC_DEBUG("WebSocket handler: rtt PING/PONG mismatch, discarding PONG");
            }
        }
 
        return;
    }
    
    if (rttRunning && !rttStart)
    {
        string error;
        Json rttParameters = Json::parse((const char*)nopoll_msg_get_payload(msg), error);
        
        if (rttParameters["cmd"].string_value().compare("rttStart") == 0)
        {
            if (rttParameters["rttRequests"].int_value())        rttRequests         = rttParameters["rttRequests"].int_value()+1;
            if (rttParameters["rttRequestTimeout"].int_value())  rttRequestTimeout   = rttParameters["rttRequestTimeout"].int_value();
            if (rttParameters["rttRequestWait"].int_value())     rttRequestWait      = rttParameters["rttRequestWait"].int_value();
            if (rttParameters["rttTimeout"].int_value())         rttTimeout          = rttParameters["rttTimeout"].int_value();
            if (rttParameters["rttPayloadSize"].int_value())     
            {
                if (rttParameters["rttPayloadSize"].int_value() > 6)
                {
                    rttPayloadSize = rttParameters["rttPayloadSize"].int_value();
                }
            }
        
            rttStart = true;
        }
    }
    
    if (uploadRunning)
    {
        uploadBytesReceived         += nopoll_msg_get_payload_size(msg);
        if (nopoll_msg_opcode(msg) != 0)
        {
            uploadHeaderReceived++;
        }
        
        return;
    }
    
    if (!uploadRunning) 
    {
        TRC_DEBUG("WebSocket handler: Message received: ");
        TRC_DEBUG((const char*)nopoll_msg_get_payload(msg));
        
        return;
    }
}


void CWsHandler::websocket_close_handler(noPollCtx *ctx, noPollConn *conn, noPollPtr user_data)
{
    TRC_DEBUG("WebSocket handler: close");
    
    rttRunning      = false;
    downloadRunning = false;
    uploadRunning   = false;
    
    nopoll_conn_flush_writes(conn, 2000, 0);
    
    nopoll_loop_stop(ctx);
}


void CWsHandler::websocket_timeout_handler(noPollCtx *ctx, noPollConn *conn)
{
    TRC_ERR("WebSocket handler: Timeout reached");
    
    TRC_DEBUG("WebSocket handler: stopped");
    
    nopoll_conn_close(conn);
    
    if (showShutdown) TRC_DEBUG("Socket: Connection Shutdown for Client IP: " + sClientIp);
}


void CWsHandler::nopoll_logging_handler(noPollCtx * ctx, noPollDebugLevel level, const char * log_msg, noPollPtr user_data)
{
    if (level == NOPOLL_LEVEL_DEBUG && ::DEBUG_NOPOLL)
    {
        TRC_DEBUG("NoPoll: " + string(log_msg));
    }
    
    if (level == NOPOLL_LEVEL_WARNING && DEBUG_NOPOLL_ERROR)
    {
        TRC_ERR("NoPoll: " + string(log_msg));
    }
        
    if (level == NOPOLL_LEVEL_CRITICAL && DEBUG_NOPOLL_CRITICAL)
    {
        TRC_CRIT("NoPoll: " + string(log_msg));
        
        nopoll_loop_stop(ctx);
    }
    
    return;
}


int CWsHandler::roundTripTime(noPollCtx *ctx, noPollConn *conn)
{
    TRC_DEBUG("WebSocket handler: Round Trip Time");
    
    rttRunning = true;
    
    unsigned long long startTime;
    unsigned long long runningTime;
    
    startTime = CTool::get_timestamp();
    
    do
    {
        runningTime = CTool::get_timestamp() - startTime;
        
        if (!rttStart)
        {
            usleep(100);
            if (((runningTime / 1000) > rttTimeout) || !rttRunning) break;
            else continue;
        }
        
        if (showRttStart)
        {
            showRttStart = false;
            TRC_DEBUG("WebSocket handler: Round Trip Time start");
        }
        
        char randomData[rttPayloadSize-6];
        CTool::randomData(randomData, rttPayloadSize-6);
        
        rttPayloadTimestamp = to_string(CTool::get_timestamp_usec());
        
        rttPayload = rttPayloadTimestamp + ";" + randomData;

        if (nopoll_conn_send_frame (conn, nopoll_true, nopoll_false, NOPOLL_PING_FRAME, rttPayloadSize, (noPollPtr)rttPayload.c_str(), 0) == -1)
        {
            rttErrors++;
            setRoundTripTimeKPIs();
            sendRoundTripTimeResponse(ctx, conn);
            rttPingSendTime     = 0;
            rttPongReceiveTime  = 0;
            rttPayloadTimestamp = "0";
            usleep(rttRequestWait * 1000); 
            continue;
        }
        
        rttPingSendTime = CTool::get_timestamp();
        
        rttRequestsSend++;
        
        TRC_DEBUG("WebSocket handler: rtt PING send");
        
        while (rttPongReceiveTime == 0)
        {
            if (((CTool::get_timestamp() - rttPingSendTime) / 1000) > rttRequestTimeout)
            {
                TRC_DEBUG("WebSocket handler: Round Trip Time Request Timeout");
                if (rttRequestsSend != 1)
                {
                    rttMissing++;
                    setRoundTripTimeKPIs();
                    sendRoundTripTimeResponse(ctx, conn);
                }

                break;
            }
            usleep(300);
        }

        if (((CTool::get_timestamp() - rttPingSendTime) / 1000) > rttRequestTimeout)
        {
            rttPingSendTime         = 0;
            rttPongReceiveTime      = 0;
            rttPayloadTimestamp     = "0";
            usleep(rttRequestWait * 2000);
            continue;
        }
        
        long long rtt = rttPongReceiveTime - rttPingSendTime;

        if (rttRequestsSend != 1)
        {
            rttVector.push_back(rtt);

            rttReplies++;
            setRoundTripTimeKPIs();
            sendRoundTripTimeResponse(ctx, conn);
        }
        
        rttPingSendTime        = 0;
        rttPongReceiveTime     = 0;
        
        usleep(rttRequestWait * 1000);
        
    } while (((runningTime  / 1000) < rttTimeout) && rttRunning && rttRequestsSend < rttRequests);

    nopoll_conn_flush_writes(conn, 2000, 0);
    
    if (rttRequestsSend == rttRequests)
    {
        usleep(6000 * 1000);
        nopoll_conn_close(conn);
        return 1;
    }
    
    if (rttRunning && ((runningTime / 1000) >= rttTimeout)) 
    {
        showShutdown = true;
        showStopped  = false;
        websocket_timeout_handler(ctx, conn);
    }
    
    return 1;
}


void CWsHandler::setRoundTripTimeKPIs()
{
    long long rttSum = 0;

    for (long long rtt : rttVector)
    {
        rttSum += rtt;
    }
    
    if (rttSum != 0)
    {
        rttAvg = rttSum / rttReplies;
        
        if ((rttVector.back() < rttMin) || rttMin == 0)
        {
            rttMin = rttVector.back();
        }
        if (rttVector.back() > rttMax)
        {
            rttMax = rttVector.back();
        }

        vector<long long> rttMedianVector = rttVector;
        sort(rttMedianVector.begin(), rttMedianVector.end());
        size_t length = rttMedianVector.size();
        
        if (length%2 == 0)
        {
            rttMed  = (rttMedianVector[length/2-1] + rttMedianVector[length/2])/2;
        }
        else
        {
            rttMed = rttMedianVector[length/2];
        }
    }
    else
    {
        rttAvg = 0;
        rttMed = 0;
        rttMin = 0;
        rttMax = 0;
    }
         
    rttMissing      = rttRequestsSend - 1 - rttReplies - rttErrors;
    rttPacketsize   = rttPayloadSize;
}


void CWsHandler::sendRoundTripTimeResponse(noPollCtx *ctx, noPollConn *conn)
{
    if (((rttRequestsSend-1)%2 == 1) || (rttRequestsSend == rttRequests))
    {
        Json rttReport = Json::object{
            {"cmd",  "rttReport"},
            {"avg",  rttAvg / 1000},
            {"med",  rttMed / 1000},  
            {"min",  rttMin / 1000},
            {"max",  rttMax / 1000},
            {"req",  rttRequestsSend - 1},
            {"rep",  rttReplies},
            {"err",  rttErrors},
            {"mis",  rttMissing},
            {"pSz",  rttPacketsize},
            {"srv",  hostname},
        };

        nopoll_conn_send_text(conn, rttReport.dump().c_str(), rttReport.dump().length());

        TRC_DEBUG("WebSocket handler: rtt report send");
    }
}


int CWsHandler::download(noPollCtx *ctx, noPollConn *conn)
{
    TRC_DEBUG("WebSocket handler: download");
    
    downloadRunning = true;
    
    vector<char>randomDataValues;
    randomDataValues.clear();
    CTool::randomData(randomDataValues, downloadRandomDataSize);

    unsigned long long index = 0;
    
    unsigned long long startTime;
    unsigned long long runningTime;
    
    int nResponse = 0;
    
    usleep(500000);
    
    startTime = CTool::get_timestamp_sec();

    do
    {
        vector<char> payload(&randomDataValues[index], &randomDataValues[index+downloadFrameSize]);
        index += downloadFrameSize;
        
        if (index > randomDataValues.size())
        {
            index = index - randomDataValues.size();
            
            vector<char> payload(&randomDataValues[index], &randomDataValues[index+downloadFrameSize]);
            index += downloadFrameSize;
        }
        
        nResponse = nopoll_conn_send_binary(conn, payload.data(), downloadFrameSize);
                
        if (nResponse <= 0)
        {
            TRC_ERR("WebSocket handler: download send: " + to_string(nResponse));
            break;
        }

        runningTime = CTool::get_timestamp_sec() - startTime;
        
    } while (nResponse > 0 && runningTime < webSocketTimeout && downloadRunning);
    
    nopoll_conn_flush_writes(conn, 2000, 0);
    
    if (downloadRunning && runningTime >= webSocketTimeout) 
    {
        showShutdown = true;
        showStopped  = false;
        websocket_timeout_handler(ctx, conn);
    }
    
    return 1;
}


int CWsHandler::upload(noPollCtx *ctx, noPollConn *conn)
{
    TRC_DEBUG("WebSocket handler: upload");
    
    uploadRunning = true;
    
    unsigned long long startTime    = 0;
    unsigned long long endTime      = CTool::get_timestamp();
    unsigned long long runningTime  = 0;
    unsigned long long currentTime  = endTime / 100000;
    
    currentTime = formatCurrentTime(endTime, currentTime);

    startTime   = CTool::get_timestamp_sec();
     
    do
    {
        if ((endTime - (currentTime * 100000)) > 500000)
        {
            string uploadBytesReceivedString = to_string(uploadBytesReceived);
            uploadBytesReceived = 0;
            string uploadHeaderReceivedString = to_string(uploadHeaderReceived);
            uploadHeaderReceived = 0;
            string currentTimeString = to_string(currentTime);
            currentTime = formatCurrentTime(endTime, currentTime);
            
            Json rtReport = Json::object{
                {"cmd", "ulReport"},
                {"time", currentTimeString},
                {"bRcv", uploadBytesReceivedString},
                {"hRcv", uploadHeaderReceivedString},
            };
            
            nopoll_conn_send_text(conn, rtReport.dump().c_str(), rtReport.dump().length());   
        }
        
        usleep(1000);
        
        endTime = CTool::get_timestamp();
        runningTime = CTool::get_timestamp_sec() - startTime;
        
    } while (runningTime < webSocketTimeout && uploadRunning);
    
    nopoll_conn_flush_writes(conn, 2000, 0);
    
    if (uploadRunning && runningTime >= webSocketTimeout) 
    {
        showShutdown = false;
        showStopped  = false;
        websocket_timeout_handler(ctx, conn);
    }
    
    return 1;
}


unsigned long long CWsHandler::formatCurrentTime(unsigned long long endTime, unsigned long long currentTime)
{
    currentTime = endTime / 100000;

    if ((currentTime % 10) < 5)
    {
        currentTime -= currentTime % 10;
    }
    else
    {
        currentTime = (currentTime - (currentTime % 10)) + 5;
    }
    
    return currentTime;
}





