/*!   
    \file wsserver.cpp
    \brief WebSocket Measurement Server
    \author zafaco GmbH <info@zafaco.de>
    \date Last update: 2017-11-24

    Copyright (C) 2016 - 2017 zafaco GmbH

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3 
    as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#define VERSION 1.1
#define __NOPOLL_PTHREAD_SUPPORT__


#include "header.h"
#include "wslistener.h"




/*--------------Global Variables--------------*/

bool DEBUG;
bool DEBUG_NOPOLL;
bool DEBUG_NOPOLL_ERROR;
bool DEBUG_NOPOLL_CRITICAL;
bool DAEMONMODE;
bool RUNNING;

int mPort       = 80;




/*--------------Forward declarations--------------*/

void        show_usage      (char* argv0);
int         start_daemon    (int nochdir, int noclose);
static void signal_handler  (int signal);
int         write_pidfile   (string sFilename);




/*--------------Beginning of Program--------------*/

int main(int argc, char** argv)
{
    ::DEBUG_NOPOLL          = false;
    ::DEBUG_NOPOLL_ERROR    = false;
    ::DEBUG_NOPOLL_CRITICAL = false;
    
    ::RUNNING               = true;

    long int opt;

    while ((opt = getopt(argc, argv, "dp:nhv")) != -1)
    {
        switch (opt)
        {
            case 'd':
                ::DAEMONMODE = true;
                break;
            case 'p':
                if (atoi(optarg) > 0 && atoi(optarg) < 65536) mPort = atoi(optarg);
                else
                {
                    cout << "Error: Invalid TCP Port: " << optarg << endl;
                    return EXIT_FAILURE;
                }
                break;
            case 'n':
                ::DEBUG_NOPOLL          = true;
                ::DEBUG_NOPOLL_ERROR    = true;
                ::DEBUG_NOPOLL_CRITICAL = true;
                break;
            case 'h':
                show_usage(argv[0]);
                return EXIT_SUCCESS;
            case 'v':
                cout << "zafaco WebSocket Measurement Server" << endl;
                cout << "Version: " << VERSION << endl;
                return EXIT_SUCCESS;
            case '?':
            default:
                cout << "Error: Unknown Argument -" << opt << endl;
                show_usage(argv[0]);
                return EXIT_FAILURE;
        }
    }

    if (::DAEMONMODE)
    {
        if (start_daemon(0, 0) < 0)
        {
            cout << "Error: Failed to create a Daemon" << endl;
            return EXIT_FAILURE;
        }
    }

    signal(SIGINT, signal_handler);

    write_pidfile("/var/run/wsserver.pid");
    
    TRC_INFO("Status: wsServer started");
    
    CWsListener *pWsListener = new CWsListener(mPort);
    
    if (pWsListener->createThread() != 0)
    {
        TRC_ERR("Error: Failure while creating WS Listener Thread");
        return EXIT_FAILURE;
    }
    
    pWsListener->waitForEnd();
    
    delete(pWsListener);
    
    TRC_INFO("Status: wsServer stopped");
    
    return EXIT_SUCCESS;
}


void show_usage(char* argv0)
{
    cout << "                                                      " << endl;
    cout << "Usage: " << argv0 << " [ options ... ]                " << endl;
    cout << "                                                      " << endl;
    cout << "  -d             - Daemon-Mode: Switch into a daemon  " << endl;
    cout << "  -p [port]      - Set WS TCP Port [default: "          << to_string(mPort)    << "]" << endl;
    cout << "  -n             - Show nopoll Debugging output       " << endl;
    cout << "  -h             - Show Help                          " << endl;
    cout << "  -v             - Show Version                       " << endl;
    cout << "                                                      " << endl;

    exit(EXIT_FAILURE);
}


int start_daemon(int nochdir, int noclose)
{
    int i = 0;
    int r = 0;

    switch (fork())
    {
        case 0: break;
        case -1: return -1;

        default: exit(EXIT_SUCCESS);
    }

    if (setsid() < 0)
    {
        return -1; 
    }

    switch (fork())
    {
        case 0: break;
        case -1: return -1;

        default: exit(EXIT_FAILURE);
    }

    if (!nochdir)
    {
        r = chdir("/");       
    }

    if (!noclose)
    {
        for (; i < 255; ++i)
        {
            close(i);
        }
            
        open("/dev/null", O_RDWR);
        r = dup(0);
        r = dup(0);
    }

    r = 0;

    return r;
}


static void signal_handler(int signal)
{
    ::RUNNING = false;
    sleep(1);
    exit(signal);
}


int write_pidfile(string sFilename)
{
    ofstream fout((sFilename).c_str());

    fout << getpid() << endl;

    fout.close();

    return 0;
}
